/*
 * A write to C0x0, turn the high bit off and write to a file.
 * ROM code is taken from Andrew Gregory's emulator.
 * Remaining code by Steve Nickolas
 */

#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <windows.h>

__declspec(dllexport) uint32_t mod_type=0x544F4C53;
__declspec(dllexport) uint8_t *mod_name="LPT Interface";
__declspec(dllexport) uint8_t *mod_pref="lpt";
__declspec(dllexport) uint8_t *mod_vers="0.01";
__declspec(dllexport) uint8_t *mod_copr="Copyright 2016 Steve Nickolas\n6502 side code by Andrew Gregory";

__declspec(dllexport) uint8_t  *lpt_preload_name="";
__declspec(dllexport) uint16_t lpt_preload_size=0;
__declspec(dllexport) uint8_t  *lpt_preload_ptr=0;

static uint8_t lptrom[256]={
 0x18, 0xB0, 0x38, 0x48, 0x8A, 0x48, 0x98, 0x48,
 0x08, 0x78, 0x20, 0x58, 0xFF, 0xBA, 0x68, 0x68,
 0x68, 0x68, 0xA8, 0xCA, 0x9A, 0x68, 0x28, 0xAA,
 0x90, 0x38, 0xBD, 0xB8, 0x05, 0x10, 0x19, 0x98,
 0x29, 0x7F, 0x49, 0x30, 0xC9, 0x0A, 0x90, 0x3B,
 0xC9, 0x78, 0xB0, 0x29, 0x49, 0x3D, 0xF0, 0x21,
 0x98, 0x29, 0x9F, 0x9D, 0x38, 0x06, 0x90, 0x7E,
 0xBD, 0xB8, 0x06, 0x30, 0x14, 0xA5, 0x24, 0xDD,
 0x38, 0x07, 0xB0, 0x0D, 0xC9, 0x11, 0xB0, 0x09,
 0x09, 0xF0, 0x3D, 0x38, 0x07, 0x65, 0x24, 0x85,
 0x24, 0x4A, 0x38, 0xB0, 0x6D, 0x18, 0x6A, 0x3D,
 0xB8, 0x06, 0x90, 0x02, 0x49, 0x81, 0x9D, 0xB8,
 0x06, 0xD0, 0x53, 0xA0, 0x0A, 0x7D, 0x38, 0x05,
 0x88, 0xD0, 0xFA, 0x9D, 0xB8, 0x04, 0x9D, 0x38,
 0x05, 0x38, 0xB0, 0x43, 0xC5, 0x24, 0x90, 0x3A,
 0x68, 0xA8, 0x68, 0xAA, 0x68, 0x4C, 0xF0, 0xFD,
 0x90, 0xFE, 0xB0, 0xFE, 0x99, 0x80, 0xC0, 0x90,
 0x37, 0x49, 0x07, 0xA8, 0x49, 0x0A, 0x0A, 0xD0,
 0x06, 0xB8, 0x85, 0x24, 0x9D, 0x38, 0x07, 0xBD,
 0xB8, 0x06, 0x4A, 0x70, 0x02, 0xB0, 0x23, 0x0A,
 0x0A, 0xA9, 0x27, 0xB0, 0xCF, 0xBD, 0x38, 0x07,
 0xFD, 0xB8, 0x04, 0xC9, 0xF8, 0x90, 0x03, 0x69,
 0x27, 0xAC, 0xA9, 0x00, 0x85, 0x24, 0x18, 0x7E,
 0xB8, 0x05, 0x68, 0xA8, 0x68, 0xAA, 0x68, 0x60,
 0x90, 0x27, 0xB0, 0x00, 0x10, 0x11, 0xA9, 0x89,
 0x9D, 0x38, 0x06, 0x9D, 0xB8, 0x06, 0xA9, 0x28,
 0x9D, 0xB8, 0x04, 0xA9, 0x02, 0x85, 0x36, 0x98,
 0x5D, 0x38, 0x06, 0x0A, 0xF0, 0x90, 0x5E, 0xB8,
 0x05, 0x98, 0x48, 0x8A, 0x0A, 0x0A, 0x0A, 0x0A,
 0xA8, 0xBD, 0x38, 0x07, 0xC5, 0x24, 0x68, 0xB0,
 0x05, 0x48, 0x29, 0x80, 0x09, 0x20, 0x2C, 0x58,
 0xFF, 0xF0, 0x03, 0xFE, 0x38, 0x07, 0x70, 0x84
};

static char inifile[1024];
static uint8_t lptfile[260];
static FILE *file;
static uint8_t crlf;

__declspec(dllexport) void lpt_write16 (uint16_t a, uint8_t v)
{
 if (!(a&0x0F))
 {
  if (!file) return;
  fprintf(file,"%c",v&0x7F);
  if (((v&0x7F)==0x0D)&&(crlf)) fprintf(file,"%c",0x0A);
 }
}

__declspec(dllexport) void lpt_write256 (uint16_t a, uint8_t v)
{
}

__declspec(dllexport) void lpt_write2048 (uint16_t a, uint8_t v)
{
}

__declspec(dllexport) uint8_t lpt_read16 (uint16_t a)
{
 return 0;
}

__declspec(dllexport) uint8_t lpt_read256 (uint16_t a)
{
 return lptrom[a&0x00FF];
}

__declspec(dllexport) uint8_t lpt_read2048 (uint16_t a)
{
 return 0;
}

__declspec(dllexport) int16_t lpt_override_wr (uint16_t a, uint8_t v)
{
 return -1; /* we don't do this. */
}

__declspec(dllexport) int16_t lpt_override_rd (uint16_t a)
{
 return -1; /* we don't do this. */
}

__declspec(dllexport) int lpt_reset (void)
{
 fflush(file);
}

__declspec(dllexport) int lpt_loop (uint32_t cycles)
{
}

__declspec(dllexport) int lpt_init (void)
{
 int l;
 char modbuf[260];
 getcwd(inifile,1010);
 strcat(inifile,"\\modapple.ini");
 l=GetPrivateProfileString("mod_lpt","LptFile","printer.txt",modbuf,260,inifile);
 if (l) strcpy(lptfile,modbuf);
 l=GetPrivateProfileString("mod_lpt","LineFeeds","Off",modbuf,260,inifile);
 if (!strcasecmp(modbuf,"Off")) crlf=0; else crlf=1;
 
 file=fopen(lptfile,"ab");
 return 0;
}

__declspec(dllexport) int lpt_deinit (void)
{
 int l;
 
 l=WritePrivateProfileString("mod_lpt","LptFile",lptfile,inifile);
 l=WritePrivateProfileString("mod_lpt","LineFeeds",crlf?"On":"Off",inifile);

 if (file) fclose(file);
}

BOOL WINAPI DialogProc(HWND hWindow, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
 int x;

 OPENFILENAME ofn;
 char browsefn[260];
 
 char *foo_p;
 long  foo_l;
 HWND  foo_c;
 
 /* wParam = what was hit */
 switch(uMsg)
 {
  case WM_INITDIALOG:
   SetWindowText(GetDlgItem(hWindow, 1), lptfile);
   if (crlf) CheckDlgButton(hWindow, 4, BST_CHECKED);
   return TRUE;
  case WM_COMMAND:
   switch (wParam)
   {
    case 2: /* ESC */
    case 6: /* Cancel button */
     EndDialog(hWindow, 0);
     DestroyWindow(hWindow);
     return TRUE;
    case 5: /* OK clicked */
     foo_c=GetDlgItem(hWindow, 1);
     foo_l=GetWindowTextLength(foo_c);
     foo_p=malloc(foo_l+1);
     GetWindowText(foo_c,foo_p,foo_l+1);
     strcpy(browsefn,foo_p);
     free(foo_p);
     if(file) fclose(file);
     strcpy(lptfile,browsefn);
     crlf=(IsDlgButtonChecked(hWindow, 4))?1:0;
     file=fopen(lptfile,"ab");

     EndDialog(hWindow, 0);
     DestroyWindow(hWindow);
     return TRUE;
    case 3: /* Browse clicked */
     *browsefn=0;
     ZeroMemory(&ofn,sizeof(ofn));
     ofn.lStructSize    =sizeof(ofn);
     ofn.hwndOwner      =hWindow;
     ofn.lpstrFile      =browsefn;
     ofn.nMaxFile       =260;
     ofn.lpstrFilter    ="Text Files (*.TXT)\0*.TXT\0All Files (*.*)\0*.*\0";
     ofn.nFilterIndex   =1;
     ofn.lpstrFileTitle =0;
     ofn.nMaxFileTitle  =0;
     ofn.lpstrInitialDir=0;
     ofn.Flags          =(OFN_PATHMUSTEXIST);
     if (GetSaveFileName(&ofn))
     {
      foo_c=GetDlgItem(hWindow, 1);
      SetWindowText(foo_c,browsefn);
     }
     return TRUE;
   }
   break;
  case WM_CLOSE:
   EndDialog(hWindow, 0);
   DestroyWindow(hWindow);
   return TRUE;
 }

 return FALSE;
}

static HINSTANCE hinstance;

__declspec(dllexport) void lpt_config (void)
{
 int e;
 
 e=DialogBox(hinstance, "DLGLPT", 0, DialogProc);
}

/*
 * Needed for DialogBox to work
 */
BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpReserved)
{
 hinstance=hinstDLL;
 return TRUE;
}
