/*
 * A basic video and keyboard driver for MoDapple.
 * 
 * * mod_type should be 0x54545950 (TTYP)
 * * mod_name should be the display name of the module
 * * mod_pref should be the name prefixed to module functions
 */

#include <stdint.h>
#include <SDL/SDL.h>
#include <windows.h>

SDL_Surface *screen;
static unsigned char scr_bitwid;

static uint16_t scanline;

__declspec(dllexport) uint32_t mod_type=0x50595454;
__declspec(dllexport) uint8_t *mod_name="SDL-LAMB";
__declspec(dllexport) uint8_t *mod_pref="ttylamb";
__declspec(dllexport) uint8_t *mod_vers="0.03";
__declspec(dllexport) uint8_t *mod_copr="Copyright 2017 Steve Nickolas";

static void ll_pset(SDL_Surface *surface, int x, int y, Uint32 pixel)
{
 register int    bpp;
 register Uint8  *addr;

 if (((unsigned) x)>640)
 {
//  printf ("WARNING: X out of bounds (%u)\n",x);
  return;
 }

 if (((unsigned) y)>480)
 {
//  printf ("WARNING: Y out of bounds (%u)\n",y);
  return;
 }

 bpp = surface->format->BytesPerPixel;
 addr = (Uint8 *)surface->pixels + y * surface->pitch + x * bpp;

 switch(bpp) 
 {
  case 1:
   *addr  = pixel;
   break;
  case 2:
   *(Uint16 *)addr = pixel;
   break;
  case 3:
   if (SDL_BYTEORDER == SDL_BIG_ENDIAN)
   {
    addr[0] = (pixel >> 16) & 0xff;
    addr[1] = (pixel >> 8) & 0xff;

    addr[2] =  pixel & 0xff;
   }
   else
   {
    addr[0] =  pixel & 0xff;
    addr[1] = (pixel >> 8) & 0xff;
    addr[2] = (pixel >> 16) & 0xff;
   }
   break;
  case 4:
   *(Uint32 *)addr = pixel;
   break;
 }
}

void pset_ready (void)
{
 if (SDL_MUSTLOCK(screen)) if (SDL_LockSurface(screen)<0) return;
}

void pset_done (void)
{
 if (SDL_MUSTLOCK(screen)) SDL_UnlockSurface(screen);
}

uint32_t qpset[480][640];
//uint8_t dirtty[480];
void pset (uint16_t x, uint16_t y, uint8_t r, uint8_t g, uint8_t b)
{
 uint32_t c;
 
 c=r;
 c<<=8;
 c+=g;
 c<<=8;
 c+=b;
/*
 if (qpset[y][x]==c) return;
 dirtty[y]=1;*/
 qpset[y][x]=c;
 
 ll_pset(screen, x, y, c);
}

void sglpset (uint16_t x, uint16_t y, uint8_t r, uint8_t g, uint8_t b)
{
 if (x>=560) return;
 pset(x+40, (y<<1)+48, r, g, b);
 pset(x+40, (y<<1)+49, r, g, b);
}

void dblpset (uint16_t x, uint16_t y, uint8_t r, uint8_t g, uint8_t b)
{
 sglpset (x<<1, y, r, g, b);
 sglpset ((x<<1)+1, y, r, g, b);
}

uint8_t gattr[16][3]={
 {0x00, 0x00, 0x00},
 {0xDD, 0x00, 0x33},
 {0x00, 0x00, 0x99},
 {0xDD, 0x22, 0xDD},
 {0x00, 0x77, 0x22},
 {0x55, 0x55, 0x55},
 {0x22, 0x22, 0xFF},
 {0x66, 0xAA, 0xFF},
 {0x88, 0x55, 0x00},
 {0xFF, 0x66, 0x00},
 {0xAA, 0xAA, 0xAA},
 {0xFF, 0x99, 0x88},
 {0x11, 0xDD, 0x00},
 {0xFF, 0xFF, 0x00},
 {0x44, 0xFF, 0x99},
 {0xFF, 0xFF, 0xFF}
};

typedef struct {
 uint8_t iiemode; /* emulating a //e */
 uint8_t store80,
         rdaux,
         wraux,
         lcaux,
         cxoverride,
         c3override,
         txt80,
         dblgrx,
         mostxt,
         iou,
         an3,
         alt[2];
 uint8_t lcread,
         lcwrite,
         lcalt;
 uint8_t nogx,
         split,
         page2,
         hires;
} IIE;

static IIE *iie;

int store80,txt80,dblgrx,mostxt;

static uint8_t vramt[4][1024];
static uint8_t vramg[4][8192];

static int nogx, split, page2, hires, ttyok;
__declspec(dllexport) uint8_t ttylamb_ttykey;

int slowkeys;

unsigned char font[768]={
 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
 0x08,0x08,0x08,0x08,0x08,0x00,0x08,0x00,
 0x14,0x14,0x14,0x00,0x00,0x00,0x00,0x00,
 0x14,0x14,0x3E,0x14,0x3E,0x14,0x14,0x00,
 0x08,0x3C,0x0A,0x1C,0x28,0x1E,0x08,0x00,
 0x06,0x26,0x10,0x08,0x04,0x32,0x30,0x00,
 0x04,0x0A,0x0A,0x04,0x2A,0x12,0x2C,0x00,
 0x08,0x08,0x08,0x00,0x00,0x00,0x00,0x00,
 0x08,0x04,0x02,0x02,0x02,0x04,0x08,0x00,
 0x08,0x10,0x20,0x20,0x20,0x10,0x08,0x00,
 0x08,0x2A,0x1C,0x08,0x1C,0x2A,0x08,0x00,
 0x00,0x08,0x08,0x3E,0x08,0x08,0x00,0x00,
 0x00,0x00,0x00,0x00,0x08,0x08,0x04,0x00,
 0x00,0x00,0x00,0x3E,0x00,0x00,0x00,0x00,
 0x00,0x00,0x00,0x00,0x00,0x00,0x08,0x00,
 0x00,0x20,0x10,0x08,0x04,0x02,0x00,0x00,
 0x1C,0x22,0x32,0x2A,0x26,0x22,0x1C,0x00,
 0x08,0x0C,0x08,0x08,0x08,0x08,0x1C,0x00,
 0x1C,0x22,0x20,0x18,0x04,0x02,0x3E,0x00,
 0x3E,0x20,0x10,0x18,0x20,0x22,0x1C,0x00,
 0x10,0x18,0x14,0x12,0x3E,0x10,0x10,0x00,
 0x3E,0x02,0x1E,0x20,0x20,0x22,0x1C,0x00,
 0x38,0x04,0x02,0x1E,0x22,0x22,0x1C,0x00,
 0x3E,0x20,0x10,0x08,0x04,0x04,0x04,0x00,
 0x1C,0x22,0x22,0x1C,0x22,0x22,0x1C,0x00,
 0x1C,0x22,0x22,0x3C,0x20,0x10,0x0E,0x00,
 0x00,0x00,0x08,0x00,0x08,0x00,0x00,0x00,
 0x00,0x00,0x08,0x00,0x08,0x08,0x04,0x00,
 0x10,0x08,0x04,0x02,0x04,0x08,0x10,0x00,
 0x00,0x00,0x3E,0x00,0x3E,0x00,0x00,0x00,
 0x04,0x08,0x10,0x20,0x10,0x08,0x04,0x00,
 0x1C,0x22,0x10,0x08,0x08,0x00,0x08,0x00,
 0x1C,0x22,0x2A,0x3A,0x1A,0x02,0x3C,0x00,
 0x08,0x14,0x22,0x22,0x3E,0x22,0x22,0x00,
 0x1E,0x22,0x22,0x1E,0x22,0x22,0x1E,0x00,
 0x1C,0x22,0x02,0x02,0x02,0x22,0x1C,0x00,
 0x1E,0x22,0x22,0x22,0x22,0x22,0x1E,0x00,
 0x3E,0x02,0x02,0x1E,0x02,0x02,0x3E,0x00,
 0x3E,0x02,0x02,0x1E,0x02,0x02,0x02,0x00,
 0x3C,0x02,0x02,0x02,0x32,0x22,0x3C,0x00,
 0x22,0x22,0x22,0x3E,0x22,0x22,0x22,0x00,
 0x1C,0x08,0x08,0x08,0x08,0x08,0x1C,0x00,
 0x20,0x20,0x20,0x20,0x20,0x22,0x1C,0x00,
 0x22,0x12,0x0A,0x06,0x0A,0x12,0x22,0x00,
 0x02,0x02,0x02,0x02,0x02,0x02,0x3E,0x00,
 0x22,0x36,0x2A,0x2A,0x22,0x22,0x22,0x00,
 0x22,0x22,0x26,0x2A,0x32,0x22,0x22,0x00,
 0x1C,0x22,0x22,0x22,0x22,0x22,0x1C,0x00,
 0x1E,0x22,0x22,0x1E,0x02,0x02,0x02,0x00,
 0x1C,0x22,0x22,0x22,0x2A,0x12,0x2C,0x00,
 0x1E,0x22,0x22,0x1E,0x0A,0x12,0x22,0x00,
 0x1C,0x22,0x02,0x1C,0x20,0x22,0x1C,0x00,
 0x3E,0x08,0x08,0x08,0x08,0x08,0x08,0x00,
 0x22,0x22,0x22,0x22,0x22,0x22,0x1C,0x00,
 0x22,0x22,0x22,0x22,0x22,0x14,0x08,0x00,
 0x22,0x22,0x22,0x2A,0x2A,0x36,0x22,0x00,
 0x22,0x22,0x14,0x08,0x14,0x22,0x22,0x00,
 0x22,0x22,0x14,0x08,0x08,0x08,0x08,0x00,
 0x3E,0x20,0x10,0x08,0x04,0x02,0x3E,0x00,
 0x3E,0x06,0x06,0x06,0x06,0x06,0x3E,0x00,
 0x00,0x02,0x04,0x08,0x10,0x20,0x00,0x00,
 0x3E,0x30,0x30,0x30,0x30,0x30,0x3E,0x00,
 0x00,0x00,0x08,0x14,0x22,0x00,0x00,0x00,
 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x7F,
 0x04,0x08,0x10,0x00,0x00,0x00,0x00,0x00,
 0x00,0x00,0x1C,0x20,0x3C,0x22,0x3C,0x00,
 0x02,0x02,0x1E,0x22,0x22,0x22,0x1E,0x00,
 0x00,0x00,0x3C,0x02,0x02,0x02,0x3C,0x00,
 0x20,0x20,0x3C,0x22,0x22,0x22,0x3C,0x00,
 0x00,0x00,0x1C,0x22,0x3E,0x02,0x3C,0x00,
 0x18,0x24,0x04,0x1E,0x04,0x04,0x04,0x00,
 0x00,0x00,0x1C,0x22,0x22,0x3C,0x20,0x1E,
 0x02,0x02,0x1E,0x22,0x22,0x22,0x22,0x00,
 0x08,0x00,0x0C,0x08,0x08,0x08,0x1C,0x00,
 0x10,0x00,0x18,0x10,0x10,0x10,0x12,0x0C,
 0x02,0x02,0x22,0x12,0x0E,0x12,0x22,0x00,
 0x0C,0x08,0x08,0x08,0x08,0x08,0x1C,0x00,
 0x00,0x00,0x36,0x2A,0x2A,0x2A,0x22,0x00,
 0x00,0x00,0x1E,0x22,0x22,0x22,0x22,0x00,
 0x00,0x00,0x1C,0x22,0x22,0x22,0x1C,0x00,
 0x00,0x00,0x1E,0x22,0x22,0x1E,0x02,0x02,
 0x00,0x00,0x3C,0x22,0x22,0x3C,0x20,0x20,
 0x00,0x00,0x3A,0x06,0x02,0x02,0x02,0x00,
 0x00,0x00,0x3C,0x02,0x1C,0x20,0x1E,0x00,
 0x04,0x04,0x1E,0x04,0x04,0x24,0x18,0x00,
 0x00,0x00,0x22,0x22,0x22,0x32,0x2C,0x00,
 0x00,0x00,0x22,0x22,0x22,0x14,0x08,0x00,
 0x00,0x00,0x22,0x22,0x2A,0x2A,0x36,0x00,
 0x00,0x00,0x22,0x14,0x08,0x14,0x22,0x00,
 0x00,0x00,0x22,0x22,0x22,0x3C,0x20,0x1E,
 0x00,0x00,0x3E,0x10,0x08,0x04,0x3E,0x00,
 0x38,0x0C,0x0C,0x06,0x0C,0x0C,0x38,0x00,
 0x08,0x08,0x08,0x08,0x08,0x08,0x08,0x08,
 0x0E,0x18,0x18,0x30,0x18,0x18,0x0E,0x00,
 0x00,0x2C,0x1A,0x00,0x00,0x00,0x00,0x00,
 0x00,0x2A,0x14,0x2A,0x14,0x2A,0x00,0x00,
};

uint16_t contab[25][80];
static void putgr(uint8_t x, uint8_t y, uint8_t c)
{
 uint8_t f,b;
 uint16_t X, Y, Z;
 
 f=c&0x0F;
 b=(c&0xF0)>>4;
 X=x;
 X*=7;
 Z=y;
 Z<<=3;
 
 for (Y=0; Y<4; Y++)
 {
  dblpset(X+0,Z+Y,gattr[f][0],gattr[f][1],gattr[f][2]);
  dblpset(X+1,Z+Y,gattr[f][0],gattr[f][1],gattr[f][2]);
  dblpset(X+2,Z+Y,gattr[f][0],gattr[f][1],gattr[f][2]);
  dblpset(X+3,Z+Y,gattr[f][0],gattr[f][1],gattr[f][2]);
  dblpset(X+4,Z+Y,gattr[f][0],gattr[f][1],gattr[f][2]);
  dblpset(X+5,Z+Y,gattr[f][0],gattr[f][1],gattr[f][2]);
  dblpset(X+6,Z+Y,gattr[f][0],gattr[f][1],gattr[f][2]);
  dblpset(X+0,Z+Y+4,gattr[b][0],gattr[b][1],gattr[b][2]);
  dblpset(X+1,Z+Y+4,gattr[b][0],gattr[b][1],gattr[b][2]);
  dblpset(X+2,Z+Y+4,gattr[b][0],gattr[b][1],gattr[b][2]);
  dblpset(X+3,Z+Y+4,gattr[b][0],gattr[b][1],gattr[b][2]);
  dblpset(X+4,Z+Y+4,gattr[b][0],gattr[b][1],gattr[b][2]);
  dblpset(X+5,Z+Y+4,gattr[b][0],gattr[b][1],gattr[b][2]);
  dblpset(X+6,Z+Y+4,gattr[b][0],gattr[b][1],gattr[b][2]);
 }
}

static void putcattr (uint8_t x, uint8_t y, uint8_t c, uint8_t a)
{
 uint16_t foff, X, Y, Z;
 uint8_t f,b;
 
 if (c==219) {putgr(x, y, a); return;}
 
 if (c&0x80) return; /* XXX */
 if (c<0x20) return; 
 
 f=a&0x0F;
 b=(a&0xF0)>>4;
 
 foff=c-0x20;
 foff<<=3;
 X=x;
 X*=7;
 Z=y;
 Z<<=3;
 
 if (txt80)
 {
  for (Y=0; Y<8; Y++)
  {
   sglpset(X+0,Z+Y,(font[foff+Y]&0x01)?gattr[f][0]:gattr[b][0],(font[foff+Y]&0x01)?gattr[f][1]:gattr[b][1],(font[foff+Y]&0x01)?gattr[f][2]:gattr[b][2]);
   sglpset(X+1,Z+Y,(font[foff+Y]&0x02)?gattr[f][0]:gattr[b][0],(font[foff+Y]&0x02)?gattr[f][1]:gattr[b][1],(font[foff+Y]&0x02)?gattr[f][2]:gattr[b][2]);
   sglpset(X+2,Z+Y,(font[foff+Y]&0x04)?gattr[f][0]:gattr[b][0],(font[foff+Y]&0x04)?gattr[f][1]:gattr[b][1],(font[foff+Y]&0x04)?gattr[f][2]:gattr[b][2]);
   sglpset(X+3,Z+Y,(font[foff+Y]&0x08)?gattr[f][0]:gattr[b][0],(font[foff+Y]&0x08)?gattr[f][1]:gattr[b][1],(font[foff+Y]&0x08)?gattr[f][2]:gattr[b][2]);
   sglpset(X+4,Z+Y,(font[foff+Y]&0x10)?gattr[f][0]:gattr[b][0],(font[foff+Y]&0x10)?gattr[f][1]:gattr[b][1],(font[foff+Y]&0x10)?gattr[f][2]:gattr[b][2]);
   sglpset(X+5,Z+Y,(font[foff+Y]&0x20)?gattr[f][0]:gattr[b][0],(font[foff+Y]&0x20)?gattr[f][1]:gattr[b][1],(font[foff+Y]&0x20)?gattr[f][2]:gattr[b][2]);
   sglpset(X+6,Z+Y,(font[foff+Y]&0x40)?gattr[f][0]:gattr[b][0],(font[foff+Y]&0x40)?gattr[f][1]:gattr[b][1],(font[foff+Y]&0x40)?gattr[f][2]:gattr[b][2]);
  }
 }
 else
 {
  for (Y=0; Y<8; Y++)
  {
   dblpset(X+0,Z+Y,(font[foff+Y]&0x01)?gattr[f][0]:gattr[b][0],(font[foff+Y]&0x01)?gattr[f][1]:gattr[b][1],(font[foff+Y]&0x01)?gattr[f][2]:gattr[b][2]);
   dblpset(X+1,Z+Y,(font[foff+Y]&0x02)?gattr[f][0]:gattr[b][0],(font[foff+Y]&0x02)?gattr[f][1]:gattr[b][1],(font[foff+Y]&0x02)?gattr[f][2]:gattr[b][2]);
   dblpset(X+2,Z+Y,(font[foff+Y]&0x04)?gattr[f][0]:gattr[b][0],(font[foff+Y]&0x04)?gattr[f][1]:gattr[b][1],(font[foff+Y]&0x04)?gattr[f][2]:gattr[b][2]);
   dblpset(X+3,Z+Y,(font[foff+Y]&0x08)?gattr[f][0]:gattr[b][0],(font[foff+Y]&0x08)?gattr[f][1]:gattr[b][1],(font[foff+Y]&0x08)?gattr[f][2]:gattr[b][2]);
   dblpset(X+4,Z+Y,(font[foff+Y]&0x10)?gattr[f][0]:gattr[b][0],(font[foff+Y]&0x10)?gattr[f][1]:gattr[b][1],(font[foff+Y]&0x10)?gattr[f][2]:gattr[b][2]);
   dblpset(X+5,Z+Y,(font[foff+Y]&0x20)?gattr[f][0]:gattr[b][0],(font[foff+Y]&0x20)?gattr[f][1]:gattr[b][1],(font[foff+Y]&0x20)?gattr[f][2]:gattr[b][2]);
   dblpset(X+6,Z+Y,(font[foff+Y]&0x40)?gattr[f][0]:gattr[b][0],(font[foff+Y]&0x40)?gattr[f][1]:gattr[b][1],(font[foff+Y]&0x40)?gattr[f][2]:gattr[b][2]);
  }
 }
}

static void gputb (uint8_t x, uint8_t y, uint8_t value)
{
 uint16_t X;
 
 if (x>40) return;
 if (y>191) return;
 
 X=x;
 X*=7;
 
 dblpset(X+0,y,(value&1)?0xFF:0x00,(value&1)?0xFF:0x00,(value&1)?0xFF:0x00);
 dblpset(X+1,y,(value&2)?0xFF:0x00,(value&2)?0xFF:0x00,(value&2)?0xFF:0x00);
 dblpset(X+2,y,(value&4)?0xFF:0x00,(value&4)?0xFF:0x00,(value&4)?0xFF:0x00);
 dblpset(X+3,y,(value&8)?0xFF:0x00,(value&8)?0xFF:0x00,(value&8)?0xFF:0x00);
 dblpset(X+4,y,(value&16)?0xFF:0x00,(value&16)?0xFF:0x00,(value&16)?0xFF:0x00);
 dblpset(X+5,y,(value&32)?0xFF:0x00,(value&32)?0xFF:0x00,(value&32)?0xFF:0x00);
 dblpset(X+6,y,(value&64)?0xFF:0x00,(value&64)?0xFF:0x00,(value&64)?0xFF:0x00);
}

static void hmark (uint16_t addr)
{
 uint16_t sixtyfours, eights, ones, x, y, value;
 
 if (addr>0x1FFF) return;
 
 if ((addr&0x7F)>119) return; /* screen hole */
 
 eights=(addr&0x7F)/40;
 ones=(addr&0x03FF)>>7;
 x=(addr&0x7F)%40;
 y=(eights<<3)|ones;

 //ones=addr>>10; /* by 1024 */
 //sixtyfours=(addr&0x7F)/40;
 //eights=(addr&0x3FF)>>7; /* by 128 */
 //x=(addr&0x7F)%40;
 //y=(sixtyfours<<6)|(eights<<3)|ones;
 
 if (split&&(y>19)) return;
 if ((x<40)&&(y<24))
 {
  uint16_t yy;
  
  for (yy=0; yy<8; yy++)
  {
   value=vramg[page2][(addr&0x03FF)+(yy<<10)];
   if ((iie->iiemode)&&(store80)) value=vramg[0][(addr&0x03FF)+(yy<<10)];
   gputb(x, (y<<3)|yy, value);
  }
 }
}

static void cputc80 (uint8_t x, uint8_t y, uint8_t value1, uint8_t value2)
{
 if ((x<40)&&(y<24)) 
 {
  if ((nogx)||(split&&(y>19)))
  {
   if (value2<0x20) putcattr(x<<1,y,value2|0x40,0xF0);
    else if (value2<0x40) putcattr(x<<1,y,value2,0xF0);
    else if (value2<0x80)
    {
     if ((iie->iiemode)&&(mostxt))
     {
      if (value2<0x80) putcattr(x<<1,y,value2,0xF0);
     }
     else
     {
      if (value2<0x60) putcattr(x<<1,y,value2,0xF0);
       else if (value2<0x80) putcattr(x<<1,y,(value2&0x3F),0xF0);
     }
    }
    else if (value2<0xA0) putcattr(x<<1,y,(value2&0x1F)|0x40,0x0F);
    else putcattr(x<<1,y,value2&0x7F,0x0F);

   if (value1<0x20) putcattr(x<<1+1,y,value1|0x40,0xF0);
    else if (value1<0x40) putcattr(x<<1+1,y,value1,0xF0);
    else if (value1<0x80)
    {
     if ((iie->iiemode)&&(mostxt))
     {
      if (value1<0x80) putcattr(x<<1+1,y,value1,0xF0);
     }
     else
     {
      if (value1<0x60) putcattr(x<<1+1,y,value1,0xF0);
       else if (value1<0x80) putcattr(x<<1+1,y,(value1&0x3F),0xF0);
     }
    }
    else if (value1<0xA0) putcattr(x<<1+1,y,(value1&0x1F)|0x40,0x0F);
    else putcattr(x<<1+1,y,value1&0x7F,0x0F);
  }
  else
  {
   if (dblgrx)
   {
    putcattr(x<<1,y,219,value2);
    putcattr(x<<1+1,y,219,value1);
   }
   else
   {
    putcattr(x<<1,y,219,value1);
   }
  }
 }
}

static void cputc40 (uint8_t x, uint8_t y, uint8_t value)
{
 if ((x<40)&&(y<24)) 
 {
  if ((nogx)||(split&&(y>19)))
  {
   if (value<0x20) putcattr(x,y,value|0x40,0xF0);
    else if (value<0x40) putcattr(x,y,value,0xF0);
    else if (value<0x80)
    {
     if ((iie->iiemode)&&(mostxt))
     {
      if (value<0x80) putcattr(x,y,value,0xF0);
     }
     else
     {
      if (value<0x60) putcattr(x,y,value,0xF0);
       else if (value<0x80) putcattr(x,y,(value&0x3F),0xF0);
     }
    }
    else if (value<0xA0) putcattr(x,y,(value&0x1F)|0x40,0x0F);
    else putcattr(x,y,value&0x7F,0x0F);
  }
  else
  {
   putcattr(x,y,219,value);
  }
 }
}

static void mark (uint16_t addr)
{
 uint8_t eights, ones, x, y, value1, value2;
 
 if (addr>0x03FF) return;
 
 value1=vramt[page2][addr];
 if ((iie->iiemode)&&(store80)) value1=vramt[0][addr];

 value2=vramt[2+page2][addr];
 if ((iie->iiemode)&&(store80)) value2=vramt[2][addr];
  
 eights=(addr&0x7F)/40;
 ones=(addr&0x03FF)>>7;
 x=(addr&0x7F)%40;
 y=(eights<<3)|ones;
  
 if ((iie->iiemode)&&(txt80)) 
 {
  cputc80(x, y, value1, value2); 
 }
 else
 {
  if ((!nogx)&&(hires))
  {
   if ((!split)||(y<20)) 
   {
    hmark(addr);
    return;
   }
  }
  cputc40(x, y, value1);
 }
 
}

__declspec(dllexport) void ttylamb_ttywrite (uint32_t al, uint8_t v)
{
 uint16_t a, d;
 uint8_t b, c;
 
 a=al&0xFFFF;
 b=(al&65536)?2:0;
 
 if ((a>=0x2000)&&(a<0x6000))
 {
  c=(a>0x3FFF)?1:0;
  d=a&0x1FFF;
  
  vramg[c|b][d]=v;

  pset_ready();
  mark(d&0x03FF);
  pset_done();
  return;
 }
 
 c=(a>2047)?1:0;
 d=a&0x03FF;
 if (a<1024) return;
 if (a>3071) return;
 
 vramt[c|b][d]=v;
 
 pset_ready();
 mark(d);
 pset_done();
}

static void ttylamb_do (uint8_t a)
{
 uint16_t t;
 
 switch (a&0x000F)
 {
  case 0:
   nogx=0;
   break;
  case 1:
   nogx=1;
   break;
  case 2:
   split=0;
   break;
  case 3:
   split=1;
   break;
  case 4:
   page2=0;
   break;
  case 5:
   page2=1;
   break;
  case 6:
   hires=0;
   break;
  case 7:
   hires=1;
   break;
 }
 if (iie->iiemode)
 {
  iie->nogx=nogx;
  iie->split=split;
  iie->page2=page2;
  iie->hires=hires;
 }
 
 pset_ready();
 for (t=0; t<1024; t++)
 {
  mark(t);
 }
 pset_done();
}

__declspec(dllexport) void ttylamb_ttyout (uint8_t b, uint8_t v)
{
 uint16_t a;
 
 a=0xC000|b;
 
 if ((a&0xFFF0)==0xC010)
 {
  ttylamb_ttykey &= 0x7F;
  return;
 }
 if ((a&0xFFF0)==0xC030)
 {
  
 }
 if ((a&0xFFF0)==0xC050)
 {
  ttylamb_do(a&0x000F);
  return;
 }
}

__declspec(dllexport) uint8_t ttylamb_ttyin (uint8_t b)
{
 uint16_t a;
 
 a=0xC000|b;
 if ((a&0xFFF0)==0xC000)
  return ttylamb_ttykey;
 if ((a&0xFFF0)==0xC010)
 {
  ttylamb_ttykey &= 0x7F;
  return ttylamb_ttykey;
 }
 if ((a&0xFFF0)==0xC030)
 {
  
 }
 if ((a&0xFFF0)==0xC050)
 {
  ttylamb_do(a&0x000F);
  return 0;
 }
 return 0;
}

static int kcapslock (void)
{
 if (SDL_GetModState() & KMOD_CAPS)
  return 1;
  return 0;
}

static int kshift (void)
{
 if (SDL_GetModState() & KMOD_SHIFT)
  return 1;
  return 0;
}

static int kctrl (void)
{
 if (SDL_GetModState() & KMOD_CTRL)
  return 1;
  return 0;
}

static int klalt (void)
{
 if (SDL_GetModState() & KMOD_LALT)
  return 1;
  return 0;
}

static int kralt (void)
{
 if (SDL_GetModState() & KMOD_RALT)
  return 1;
  return 0;
}

__declspec(dllexport) void ttylamb_ttydeinit(void)
{
 SDL_Quit();
}

static uint8_t flop4080;

__declspec(dllexport) uint8_t ttylamb_ttyiie (void)
{
 uint16_t t;
 
 if (iie->iiemode)
 {
  if (flop4080!=iie->txt80)
  {
   flop4080=iie->txt80;
  }
 }
 txt80=iie->txt80;
 store80=iie->store80;
 mostxt=iie->mostxt;
 dblgrx=iie->dblgrx;

 pset_ready();
 for (t=0; t<1024; t++)
 {
  mark(t);
 }
 pset_done();
}

__declspec(dllexport) void ttylamb_ttyinit(IIE *main_iie)
{
 int t;
 
 flop4080=0;
 slowkeys=0;
 scanline=0;

 iie=main_iie;
 if (iie->iiemode)
 {
  iie->alt[0]=iie->alt[1]=0;
 }
 
// for (t=0;t<480;t++) dirtty[t]=1;
 
 if (SDL_Init(SDL_INIT_VIDEO) < 0)
 {
  fprintf(stderr,"Error initializing Video: %s\n",SDL_GetError());
  exit(1);
 }

 SDL_WM_SetCaption ("SDL Terminal (LAMB)", "Test Icon");
 SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY / 2, SDL_DEFAULT_REPEAT_INTERVAL);

 scr_bitwid=24;

 screen=SDL_SetVideoMode(640,480,scr_bitwid,SDL_SWSURFACE|SDL_HWPALETTE);

 if (!screen)
 {
  fprintf(stderr,"Error initializing Window: %s\n",SDL_GetError());
  exit(1);
 }

 ttyok=1;
 
 for (t=0; t<25; t++) memset(contab[t],0xFF,160);
}

void twink (int foo)
{
 Beep(foo?500:1000,100);
}

static int half;

__declspec(dllexport) int ttylamb_ttyloop(void)
{
 SDL_Event event;
 uint32_t key, keystrobe;
 uint16_t iscan;
 
 if (iie->iiemode)
 {
  iie->alt[0]=(klalt()<<7);
  iie->alt[1]=(kralt()<<7);
 }

 keystrobe=0;
 scanline++;
 while (scanline>261)
 {
  scanline-=262;
  half=!half;
 }

 iscan=(scanline<<1)|half;
 
 if (scanline<240)
 {
//  if (dirtty[iscan])
//  {
   SDL_UpdateRect(screen,0,iscan,640,1);
//   dirtty[iscan]=0;
//  }
 }

 while (SDL_PollEvent(&event))
 {
  switch (event.type)
  {
   case SDL_QUIT:
    return -1;
   case SDL_KEYUP:
    switch (event.key.keysym.sym)
    {
     case SDLK_F12:
      if (kctrl()) return 1;
    }
    break;
   case SDL_KEYDOWN:
    switch (event.key.keysym.sym)
    {
     case SDLK_F1: if (!(SDL_GetModState()&(KMOD_LALT|KMOD_RALT|KMOD_CTRL|KMOD_SHIFT))) return 11; break;
     case SDLK_F2: if (!(SDL_GetModState()&(KMOD_LALT|KMOD_RALT|KMOD_CTRL|KMOD_SHIFT))) return 12; break;
     case SDLK_F3: if (!(SDL_GetModState()&(KMOD_LALT|KMOD_RALT|KMOD_CTRL|KMOD_SHIFT))) return 13; break;
     case SDLK_F4: if (klalt()||kralt()) return -1;
                   if (!(SDL_GetModState()&(KMOD_LALT|KMOD_RALT|KMOD_CTRL|KMOD_SHIFT))) return 14; break;
     case SDLK_F5: if (!(SDL_GetModState()&(KMOD_LALT|KMOD_RALT|KMOD_CTRL|KMOD_SHIFT))) return 15; break;
     case SDLK_F6: if (!(SDL_GetModState()&(KMOD_LALT|KMOD_RALT|KMOD_CTRL|KMOD_SHIFT))) return 16; break;
     case SDLK_F7: if (!(SDL_GetModState()&(KMOD_LALT|KMOD_RALT|KMOD_CTRL|KMOD_SHIFT))) return 17; break;
     case SDLK_F8: if (!(SDL_GetModState()&(KMOD_LALT|KMOD_RALT|KMOD_CTRL|KMOD_SHIFT))) return 10; break;
     
     case SDLK_F9: return 6502;

     case 8:
       if (kctrl()) key=127; else key=8;
       keystrobe=1;
       break;
     case 9:  key=9;   keystrobe=1; break;
     case 13: key=13;  keystrobe=1; break;
     case 27: key=27;  keystrobe=1; break;
     case 32: key=32;  keystrobe=1; break;
     case 33: key=33;  keystrobe=1; break;
     case 34: key=34;  keystrobe=1; break;
     case 35: key=35;  keystrobe=1; break;
     case 36: key=36;  keystrobe=1; break;
     case 38: key=38;  keystrobe=1; break;
     case 39: key=39;  keystrobe=1;
              if (kshift()) key=34; 
      break;
     case 40: key=40;  keystrobe=1; break;
     case 41: key=41;  keystrobe=1; break;
     case 42: key=42;  keystrobe=1; break;
     case 43: key=43;  keystrobe=1; break;
     case 44: key=44;  keystrobe=1;
              if (kshift()) key='<';
      break;
     case 45: key=45;  keystrobe=1;
              if (kshift()) key='_';
      break;
     case 46: key=46;  keystrobe=1;
              if (kshift()) key='>';
      break;
     case 47: key=47;  keystrobe=1;
              if (kshift()) key='?';
      break;
     case 48: key=48;  keystrobe=1;
              if (kshift()) key=')';
      break;
     case 49: key=49;  keystrobe=1;
              if (kshift()) key='!';
      break;
     case 50: key=50;  keystrobe=1;
              if (kshift()) key='@';
              if (kctrl()) key=0;
      break;
     case 51: key=51;  keystrobe=1;
              if (kshift()) key='#';
      break;
     case 52: key=52;  keystrobe=1;
              if (kshift()) key='$';
      break;
     case 53: key=53;  keystrobe=1;
              if (kshift()) key='%';
      break;
     case 54: key=54;  keystrobe=1;
              if (kshift()) key='^';
              if (kctrl()) key=30;
      break;
     case 55: key=55;  keystrobe=1;
              if (kshift()) key='&';
      break;
     case 56: key=56;  keystrobe=1;
              if (kshift()) key='*';
      break;
     case 57: key=57;  keystrobe=1;
              if (kshift()) key='(';
      break;
     case 58: key=58;  keystrobe=1; break;
     case 59: key=59;  keystrobe=1;
              if (kshift()) key=':';
      break;
     case 60: key=60;  keystrobe=1;
              if (kshift()) key='<';
      break;
     case 61: key=61;  keystrobe=1;
              if (kshift()) key='+';
      break;
     case 62: key=62;  keystrobe=1;
              if (kshift()) key='>';
      break;
     case 63: key=63;  keystrobe=1; break;
     case 64: key=64;  keystrobe=1; break;
     case 91: key=91;  keystrobe=1;
              if (kshift()) key=123;
              if (kctrl()) key=27;
      break;
     case 92: key=92;  keystrobe=1;
              if (kshift()) key=124;
              if (kctrl()) key=28;
      break;
     case 93: key=93;  keystrobe=1;
              if (kshift()) key=125;
              if (kctrl()) key=29;
      break;
     case 94: key=94;  keystrobe=1;
              if (kshift()) key=126;
              if (kctrl()) key=30;
      break;
     case 95: key=95;  keystrobe=1;
              if (kctrl()) key=31;
      break;
     case 96: key=96;  keystrobe=1;
              if (kshift()) key=126;
      break;
     case 65: case 66: case 67: case 68: case 69: case 70: case 71:
     case 72: case 73: case 74: case 75: case 76: case 77: case 78:
     case 79: case 80: case 81: case 82: case 83: case 84: case 85:
     case 86: case 87: case 88: case 89: case 90:
       key=event.key.keysym.sym+32;
       if (kshift()) key=(event.key.keysym.sym)/*+32*/;
       if (kctrl())  key=(event.key.keysym.sym)-64;
       keystrobe=1;
       break;
     case 97: case 98: case 99:
     case 100: case 101: case 102: case 103: case 104: case 105: case 106:
     case 107: case 108: case 109: case 110: case 111: case 112: case 113:
     case 114: case 115: case 116: case 117: case 118: case 119: case 120:
     case 121: case 122:
       key=event.key.keysym.sym-32;
       if (kshift()) key=(event.key.keysym.sym)/*-32*/;
       if (kctrl())  key=(event.key.keysym.sym)-96;
       keystrobe=1;
       break;
     case SDLK_UP:    key=11; keystrobe=1; break;
     case SDLK_DOWN:  key=10; keystrobe=1; break;
     case SDLK_LEFT:  key=8;  keystrobe=1; break;
     case SDLK_RIGHT: key=21; keystrobe=1; break;
    }
    break;
  }
 }

 if (kcapslock())
 {
  if ((key>'@')&&(key<'[')) key+=32; else if ((key>'`')&&(key<'{')) key-=32;
 }

 if (keystrobe)
  ttylamb_ttykey=key|0x80;

 return 0;
}
